/* Pavel Surynek, 2008 - 2009 */
/* bibox_main.cpp */

/*
   The BIBOX Algorithm

   This is an implementation of the multi-robot path planning algorithm for biconnected environments with
   at least two free places. The code is not commented however it is simple enough to be understood
   without comments.

   The problem is given as an undirected graph, where each vertex of the graph is either free or
   occupied by a robot. A robot can move to a neighboring unoccupied vertex supposed that no other
   robot is simultaneously entering the target vertex. The task is to move robots preserving this
   condition to their final positions.

   This program is an implementation of the algorithm for solving the problem where the graph is
   biconnected and threre are at least two unoccupied vertices in the graph. The program runs in
   polynomial time with respect to the size of the input graph. More precisely it runs in O(|V|^3)
   where G=(V,E) is the input biconnected graph.
 */

#include "bibox.h"
#include "obox.h"

typedef enum {
	BIBOX,
	BIBOX_MIT,
	BIBOX_THETA_2,
	BIBOX_THETA_3,
	MIT
} METHOD_TYPE;

typedef enum {
	RAW,
	INVERSE,
	REDUNDANT,
	LONG
} SHORTENING_TYPE;

void test_grid_shortening(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	int Size_X;
	int Size_Y;
	int Robots;

	int Seed;

	for(int round = 1; round < 4; ++round)
	{
		switch (round)
		{
		case 1:
		{
			Size_X = 4;
			Size_Y = 4;
			Robots = Size_X * Size_Y - 2;
			Seed = 121;
			break;
		}
		case 2:
		{
			Size_X = 8;
			Size_Y = 8;
			Robots = Size_X * Size_Y - 2;
			Seed = 121;
			break;
		}
		case 3:
		{
			Size_X = 16;
			Size_Y = 16;
			Robots = Size_X * Size_Y - 2;
			Seed = 121;
			break;
		}
		case 4:
		{
			Size_X = 32;
			Size_Y = 32;
			Robots = Size_X * Size_Y - 2;
			Seed = 121;
			break;
		}
		case 5:
		{
			Size_X = 64;
			Size_Y = 64;
			Robots = Size_X * Size_Y - 2;
			Seed = 121;
			break;
		}
		default:
		{
			break;
		}
		}

		char prefix[1024];
		char suffix[1024];

		switch (method_type)
		{
		case BIBOX:
		{
			sprintf(prefix, "bibox_");
			break;
		}
		case BIBOX_MIT:
		{
			sprintf(prefix, "bibox_mit_");
			break;
		}
		case BIBOX_THETA_2:
		{
			sprintf(prefix, "bibox_theta_2_");
			break;
		}
		case BIBOX_THETA_3:
		{
			sprintf(prefix, "bibox_theta_3_");
			break;
		}
		case MIT:
		{
			sprintf(prefix, "mit_");
			break;
		}
		default:
		{
			break;
		}
		}

		switch (shortening_type)
		{
		case RAW:
		{
			sprintf(suffix, "grid_erasing_%dx%d_raw", Size_X, Size_Y);
			break;
		}
		case INVERSE:
		{
			sprintf(suffix, "grid_erasing_%dx%d_inverse", Size_X, Size_Y);
			break;
		}
		case REDUNDANT:
		{
			sprintf(suffix, "grid_erasing_%dx%d_redundant", Size_X, Size_Y);
			break;
		}
		case LONG:
		{
			sprintf(suffix, "grid_erasing_%dx%d_long", Size_X, Size_Y);
			break;
		}
		default:
		{
			break;
		}
		}


		char filename[1024];
		sprintf(filename, "%s%s", prefix, suffix);

		FILE *fw;
		fw = fopen(filename, "w");
/*
		if (Robots > 25)
		{
			Robots = 25;
		}
*/	
		for (int r = 1; r <= Robots; ++r)
		{
			clock_t begin_1, end_1;	
			clock_t begin_2, end_2;	
		
			GRAPH grid1, grid2, grid3;
			generate_grid(Size_X, Size_Y, &grid1);
			generate_grid(Size_X, Size_Y, &grid2);
			generate_grid(Size_X, Size_Y, &grid3);
			
			generate_grid_random(r, Seed, &grid1);
			generate_grid_random(r, Seed, &grid2);
			generate_grid_random(r, Seed, &grid3);

			begin_1 = clock();   

			SOLUTION_DATABASE database;
			FINE_SOLUTION solution, filtered_solution;

			printf("Solving ...\n");
			fprintf(fw, "Robots:%d\n", r);
			fprintf(fw, "Size:%d x %d\n", Size_X, Size_Y);

			bool solved;

			switch (method_type)
			{
			case BIBOX:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, false, false, false);
				break;
			}
			case BIBOX_MIT:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, true);
				break;
			}
			case BIBOX_THETA_2:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, false);
				break;
			}
			case BIBOX_THETA_3:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);
				break;
			}
			case MIT:
			{
				solved = solve_slider_obox_theta(&grid1, database, solution, false, false);
				break;
			}
			default:
			{
				solved = false;
				break;
			}
			}
// Bibox_theta 2
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, false);

// Bibox_theta 3
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);

// Bibox_mit 
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, true);

// Bibox
//			solved = solve_slider_bibox_theta(&grid1, database, solution, false, false, false);

// Mit
//			solved = solve_slider_obox_theta(&grid1, database, solution, false, false);

			end_1 = clock();
			begin_2 = clock();

			filter_fine_solution(&grid2, solution, filtered_solution);

			switch (shortening_type)
			{
			case RAW:
			{
				break;
			}
			case INVERSE:
			{
				erase_trivial_inverses(filtered_solution);
				break;
			}
			case REDUNDANT:
			{
				erase_independent_inverses(filtered_solution);
				break;
			}
			case LONG:
			{
				shorten_solution(filtered_solution, &grid3);
				break;
			}
			default:
			{
				break;
			}
			}

//			erase_trivial_inverses(filtered_solution);
//			erase_independent_inverses(filtered_solution);
//			shorten_solution(filtered_solution, &grid3);
//			execute_fine_solution(filtered_solution, &grid3);

			parallelize_fine_solution_multirobotic(filtered_solution);
			end_2 = clock();

			fprintf(fw, "Solution size:%d\n", solution.size());
			fprintf(fw, "Filtered solution size:%d\n", filtered_solution.size());
			fprintf(fw, "Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
			fprintf(fw, "Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
			fprint_step_parallelism(fw, filtered_solution);
		
			fprintf(fw, "Solving time:%.3f\n", (end_1 - begin_1) / (double)CLOCKS_PER_SEC);
			fprintf(fw, "Filtering time:%.3f\n", (end_2 - begin_2) / (double)CLOCKS_PER_SEC);

			fflush(fw);
		}
	}
}


void test_exchange_grids(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	clock_t begin, end;
/*
	GRAPH biconnected1, biconnected2;

	srand(123);
	generate_biconnected(5, 5, 10, 2, &biconnected1);

	srand(123);
	generate_biconnected(5, 5, 10, 2, &biconnected2);

	biconnected1.print();
*/

/* Grid 1 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(10, 6, &grid1);
	generate_grid(10, 6, &grid2);
	generate_grid(10, 6, &grid3);

	generate_grid_exchange(10, 6, 2, 0, 3, 5, 8, 0, &grid1);
	generate_grid_exchange(10, 6, 2, 0, 3, 5, 8, 0, &grid2);
	generate_grid_exchange(10, 6, 2, 0, 3, 5, 8, 0, &grid3);

	grid1.print();
*/

/* Grid 2 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(16, 10, &grid1);
	generate_grid(16, 10, &grid2);
	generate_grid(16, 10, &grid3);

	generate_grid_exchange(16, 10, 2, 0, 3, 9, 14, 0, &grid1);
	generate_grid_exchange(16, 10, 2, 0, 3, 9, 14, 0, &grid2);
	generate_grid_exchange(16, 10, 2, 0, 3, 9, 14, 0, &grid3);

	grid1.print();
*/

/* Grid 3 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(16, 10, &grid1);
	generate_grid(16, 10, &grid2);
	generate_grid(16, 10, &grid3);

	generate_grid_exchange(16, 10, 2, 0, 4, 9, 13, 0, &grid1);
	generate_grid_exchange(16, 10, 2, 0, 4, 9, 13, 0, &grid2);
	generate_grid_exchange(16, 10, 2, 0, 4, 9, 13, 0, &grid3);

	grid1.print();
*/

/* Grid 4 */

	GRAPH grid1, grid2, grid3;
	generate_grid(20, 16, &grid1);
	generate_grid(20, 16, &grid2);
	generate_grid(20, 16, &grid3);

	generate_grid_exchange(20, 16, 2, 0, 4, 15, 16, 0, &grid1);
	generate_grid_exchange(20, 16, 2, 0, 4, 15, 16, 0, &grid2);
	generate_grid_exchange(20, 16, 2, 0, 4, 15, 16, 0, &grid3);

	grid1.print();


/* Grid 5 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(20, 16, &grid1);
	generate_grid(20, 16, &grid2);
	generate_grid(20, 16, &grid3);

	generate_grid_exchange(20, 16, 2, 0, 6, 15, 14, 0, &grid1);
	generate_grid_exchange(20, 16, 2, 0, 6, 15, 14, 0, &grid2);
	generate_grid_exchange(20, 16, 2, 0, 6, 15, 14, 0, &grid3);

	grid1.print();
*/
/* Grid 6 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(10, 6, &grid1);
	generate_grid(10, 6, &grid2);
	generate_grid(10, 6, &grid3);

	generate_grid_exchange(10, 6, 2, 0, 5, 5, 6, 0, &grid1);
	generate_grid_exchange(10, 6, 2, 0, 5, 5, 6, 0, &grid2);
	generate_grid_exchange(10, 6, 2, 0, 5, 5, 6, 0, &grid3);

	grid1.print();
*/
/*	
	printf("Solving ...\n");

	bool solved;
	solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);
*/
/*
	grid2.print();
	grid1.print();
*/
	begin = clock();   

	SOLUTION_DATABASE database;
	FINE_SOLUTION solution, filtered_solution;

	printf("Solving ...\n");

	bool solved;

// Bibox_theta 2
//	solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, false);

// Bibox_theta 3
//	solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);

// Bibox_mit 
//	solved = solve_slider_bibox_theta(&biconnected1, database, solution, true, false, true);

// Bibox
//	solved = solve_slider_bibox_theta(&biconnected1, database, solution, false, false, false);
//	solved = solve_slider_bibox_theta(&grid1, database, solution, false, false, false);

//	printf("Finished ...\n");

//	filter_fine_solution(&biconnected2, solution, filtered_solution);

	filter_fine_solution(&grid2, solution, filtered_solution);
//	erase_trivial_inverses(filtered_solution);
//	erase_independent_inverses(filtered_solution);
	shorten_solution(filtered_solution, &grid3);
	parallelize_fine_solution_multirobotic(filtered_solution);

	print_fine_solution(filtered_solution);

	printf("Solution size:%d\n", solution.size());
	printf("Filtered solution size:%d\n", filtered_solution.size());
	printf("Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
	printf("Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
	print_step_parallelism(filtered_solution);

	end = clock();
	printf("Time:%.3f\n", (end - begin) / (double)CLOCKS_PER_SEC);

/*
	int dur = calc_solution_duration(filtered_solution);
	for (int k = 0; k <= dur; ++k)
	{
		printf("Step:%d\n", k);
		show_state_in_grid(30, 20, &grid3, k, filtered_solution);
		printf("\n");
	}
*/
}


void test_initial(void)
{
//	exit(0);
/*
	g.m_cycles.push_back(CYCLE());
	CYCLE *cycle = &g.m_cycles.back();

	cycle->m_x_id = 3; // Same 1
	cycle->m_y_id = 5; // Same 1 + 2

	g.add_vertex(1, 0, 6, 6);
	cycle->m_vertices.push_back(1);
	cycle->m_loop.push_back(1);
	g.add_vertex(2, 0, 5, 5);
	cycle->m_vertices.push_back(2);
	cycle->m_loop.push_back(2);
	g.add_vertex(3, 0, 1, 3);  // Same 1
	cycle->m_vertices.push_back(3);
	cycle->m_loop.push_back(3);
	g.add_vertex(4, 0, 2, 4);
	cycle->m_vertices.push_back(4);
	cycle->m_loop.push_back(4);
	g.add_vertex(5, 0, 8, 11);
	cycle->m_vertices.push_back(5);
	cycle->m_loop.push_back(5);
	g.add_vertex(6, 0, 7, 12);
	cycle->m_vertices.push_back(6);
	cycle->m_loop.push_back(6);
	g.add_vertex(7, 0, -1, 7);
	cycle->m_vertices.push_back(7);
	cycle->m_loop.push_back(7);
	g.add_vertex(8, 0, 0, 8);
	cycle->m_vertices.push_back(8);
	cycle->m_loop.push_back(8);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 4;
	cycle->m_y_id = 8;
	g.add_vertex(9, 1, 3, 1);
	cycle->m_vertices.push_back(9);
	cycle->m_loop.push_back(9);
	g.add_vertex(10, 1, 4, 2);
	cycle->m_vertices.push_back(10);
	cycle->m_loop.push_back(10);
	g.add_vertex(11, 1, 11, 0);
	cycle->m_vertices.push_back(11);
	cycle->m_loop.push_back(11);
	g.add_vertex(12, 1, 12, -1);
	cycle->m_vertices.push_back(12);
	cycle->m_loop.push_back(12);

	cycle->m_vertices.push_back(8);
	cycle->m_vertices.push_back(4);

	g.add_edge(1, 2, 0);
	g.add_edge(2, 3, 0);
	g.add_edge(3, 4, 0);
	g.add_edge(4, 5, 0);
	g.add_edge(5, 6, 0);
	g.add_edge(6, 7, 0);
	g.add_edge(7, 8, 0);
	g.add_edge(8, 1, 0);

	g.add_edge(4, 9, 1);
	g.add_edge(9, 10, 1);
	g.add_edge(10, 11, 1);
	g.add_edge(11, 12, 1);
	g.add_edge(12, 8, 1);
*/
/*
	g.m_cycles.push_back(CYCLE());
	CYCLE *cycle = &g.m_cycles.back();

	cycle->m_x_id = 2;
	cycle->m_y_id = 4; // Same 1

	g.add_vertex(1, 0, -1, 5);
	cycle->m_vertices.push_back(1);
	cycle->m_loop.push_back(1);
	g.add_vertex(2, 0, 0, 15);
	cycle->m_vertices.push_back(2);
	cycle->m_loop.push_back(2);
	g.add_vertex(3, 0, 1, 2);
	cycle->m_vertices.push_back(3);
	cycle->m_loop.push_back(3);
	g.add_vertex(4, 0, 2, 4); // Same 1
	cycle->m_vertices.push_back(4);
	cycle->m_loop.push_back(4);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 3;
	cycle->m_y_id = 4;
	g.add_vertex(5, 1, 3, 1);
	cycle->m_vertices.push_back(5);
	cycle->m_loop.push_back(5);
	g.add_vertex(6, 1, 4, 3);
	cycle->m_vertices.push_back(6);
	cycle->m_loop.push_back(6);

	cycle->m_vertices.push_back(4);
	cycle->m_vertices.push_back(3);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 5;
	cycle->m_y_id = 6;
	g.add_vertex(7, 2, 5, 6);
	cycle->m_vertices.push_back(7);
	cycle->m_loop.push_back(7);
	g.add_vertex(8, 2, 6, 13);
	cycle->m_vertices.push_back(8);
	cycle->m_loop.push_back(8);

	cycle->m_vertices.push_back(6);
	cycle->m_vertices.push_back(5);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 1;
	cycle->m_y_id = 8;
	g.add_vertex(9, 3, 15, 9);
	cycle->m_vertices.push_back(9);
	cycle->m_loop.push_back(9);
	g.add_vertex(10, 3, 14, 12);
	cycle->m_vertices.push_back(10);
	cycle->m_loop.push_back(10);
	g.add_vertex(11, 3, 13, 11);
	cycle->m_vertices.push_back(11);
	cycle->m_loop.push_back(11);
	g.add_vertex(12, 3, 12, 10);
	cycle->m_vertices.push_back(12);
	cycle->m_loop.push_back(12);
	g.add_vertex(13, 3, 11, 14);
	cycle->m_vertices.push_back(13);
	cycle->m_loop.push_back(13);
	g.add_vertex(14, 3, 10, 0);
	cycle->m_vertices.push_back(14);
	cycle->m_loop.push_back(14);
	g.add_vertex(15, 3, 9, -1);
	cycle->m_vertices.push_back(15);
	cycle->m_loop.push_back(15);

	cycle->m_vertices.push_back(8);
	cycle->m_vertices.push_back(6);
	cycle->m_vertices.push_back(4);
	cycle->m_vertices.push_back(1);

	g.add_edge(1, 2, 0);
	g.add_edge(2, 3, 0);
	g.add_edge(3, 4, 0);
	g.add_edge(4, 1, 0);
	g.add_edge(3, 5, 1);
	g.add_edge(5, 6, 1);
	g.add_edge(6, 4, 1);
	g.add_edge(5, 7, 2);
	g.add_edge(7, 8, 2);
	g.add_edge(8, 6, 2);
	g.add_edge(1, 9, 3);
	g.add_edge(9, 10, 3);
	g.add_edge(10, 11, 3);
	g.add_edge(11, 12, 3);
	g.add_edge(12, 13, 3);
	g.add_edge(13, 14, 3);
	g.add_edge(14, 15, 3);
	g.add_edge(15, 8, 3);
*/

//	FILE *s;
//	s = fopen("large_problems.txt", "w");
/*
	FILE *b_lpg, *b_sgplan, *b_stan, *b_ipp, *b_max, *b_sat;
	b_lpg = fopen("batch_LPG_lenght=3_free=90.bat", "w");
	b_sgplan = fopen("batch_SGPLAN_lenght=3_free=90.bat", "w");
	b_stan = fopen("batch_STAN_lenght=3_free=90.bat", "w");
	b_ipp = fopen("batch_IPP_lenght=3_free=90.bat", "w");
	b_max = fopen("batch_MAX_lenght=3_free=90.bat", "w");
	b_sat = fopen("batch_SAT_lenght=3_free=90.bat", "w");
*/
}


void test_biconnected(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	clock_t begin_1, end_1;
	clock_t begin_2, end_2;

	int ID = 0;

	int free = 2;

	int Seed;
	int init;
	int iter;
	int length;
/*
	int Seed = 25;
	int init = 5;
	int iter = 43;
	int length = 2;
	// |V|=91
	*/
/*
	int Seed = 17;
	int init = 5;
	int iter = 42;
	int length = 3;
	// |V|=89
*/
/*
	int Seed = 16;
	int init = 7;
	int iter = 38;
	int length = 4;
	// |V|=91
*/
/*
	int Seed = 1013;
	int init = 7;
	int iter = 32;
	int length = 5;
	// |V|=89
*/
/*
	int Seed = 1214;
	int init = 7;
	int iter = 30;
	int length = 6;
	// |V|=89
*/
/*
	int Seed = 1227;
	int init = 7;
	int iter = 26;
	int length = 7;
	// |V|=90
*/
/*
	int Seed = 1254;
	int init = 7;
	int iter = 22;
	int length = 8;
	// |V|=91
*/
/*
	int Seed = 1267;
	int init = 7;
	int iter = 20;
	int length = 9;
	// |V|=90
*/
/*
	int Seed = 1267;
	int init = 7;
	int iter = 18;
	int length = 10;
	// |V|=90
*/
/*
	int Seed = 1279;
	int init = 7;
	int iter = 16;
	int length = 11;
	// |V|=91
*/
/*
	int Seed = 1301;
	int init = 7;
	int iter = 14;
	int length = 12;
	// |V|=91
*/
/*
	int Seed = 1306;
	int init = 7;
	int iter = 13;
	int length = 13;
	// |V|=91
*/
/*
	int Seed = 1317;
	int init = 9;
	int iter = 12;
	int length = 14;
	// |V|=91
*/
/*
	int Seed = 1323;
	int init = 9;
	int iter = 10;
	int length = 15;
	// |V|=91
*/
/*
	int Seed = 1342;
	int init = 11;
	int iter = 9;
	int length = 16;
	// |V|=89
*/

//	for (iter = 32; iter >= 32; --iter)
	{
//		for (length = 5; length <= 7; ++length)
		{
			for (free = 2; free <= 100; free += 1)
			{
				int cnt = 0;
//				for (cnt = 0; cnt < 10; ++cnt)
				{
					printf("ID:%d (%d,%d,%d)\n", ID, iter, length, free);
					++ID;

					SOLUTION_DATABASE database;
					FINE_SOLUTION solution, filtered_solution;

					MOVE_COUNT = 0;
					STEP_COUNT = 0;
					GRAPH g, h, ch, hv;

					char name[1024];
					char filename[1024];
					
					srandom(Seed);
					generate_biconnected(init, iter, length, free, &g);

//					free = 9 * g.m_vertices.size() / 10;

					if (free < 2)
					{
						free = 2;
					}

					srandom(Seed);
					generate_biconnected(init, iter, length, free, &h);

					srandom(Seed);
					generate_biconnected(init, iter, length, free, &ch);

					srandom(Seed);
					generate_biconnected(init, iter, length, free, &hv);

					printf("Graph |V|:%d\n", g.m_vertices.size());

					if (h.m_vertices.size() > free)
					{
						begin_1 = clock();   

						bool solved;

						switch (method_type)
						{
						case BIBOX:
						{
							solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
							break;
						}
						case BIBOX_MIT:
						{
							solved = solve_slider_bibox_theta(&h, database, solution, true, false, true);
							break;
						}
						case BIBOX_THETA_2:
						{
							solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
							break;
						}
						case BIBOX_THETA_3:
						{
							solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
							break;
						}
						case MIT:
						{
							solved = solve_slider_obox_theta(&h, database, solution, false, false);
							break;
						}
						default:
						{
							solved = false;
							break;
						}
						}
// Bibox_theta 2
// 					        solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);

// Bibox_theta 3
//						solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);

// Bibox
//						solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);

// Mit
//						solved = solve_slider_obox_theta(&h, database, solution, false, false);

// Bibox_mit
//						solved = solve_slider_bibox_theta(&h, database, solution, true, false, true);

						end_1 = clock();
						begin_2 = clock();

//					        h.print();
//						hv.print();

						filter_fine_solution(&ch, solution, filtered_solution);

//						erase_trivial_inverses(filtered_solution);
//						erase_long_inverses(filtered_solution);
//						erase_independent_inverses(filtered_solution);
//						shorten_solution(filtered_solution, &hv);

						end_2 = clock();

						parallelize_fine_solution_multirobotic(filtered_solution);

						printf("Solution size:%d\n", solution.size());
						printf("Filtered solution size:%d\n", filtered_solution.size());
						printf("Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
						printf("Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
						print_step_parallelism(filtered_solution);
						//						print_fine_solution(filtered_solution);

						//						exit(0);

					}
					else
					{
						--cnt;
					}
				}
			}
			exit(0);
		}
	}
}


void test_biconnected_shortening(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	clock_t begin_1, end_1;
	clock_t begin_2, end_2;

	int Seed;
	int init;
	int iter;
	int length;

	for (length = 2; length <= 16; ++length)
	{
		int free = 2;
		int ID = 0;

		switch (length)
		{
		case 2:
		{
			Seed = 25;
			init = 5;
			iter = 43;
			length = 2;
			// |V|=91
			break;
		}
		case 3:
		{
			Seed = 17;
			init = 5;
			iter = 42;
			length = 3;
			// |V|=89	
			break;
		}
		case 4:
		{
			Seed = 16;
			init = 7;
			iter = 38;
			length = 4;
			// |V|=91	
			break;
		}
		case 5:
		{
			Seed = 1013;
			init = 7;
			iter = 32;
			length = 5;
			// |V|=89	
			break;
		}
		case 6:
		{
			Seed = 1214;
			init = 7;
			iter = 30;
			length = 6;
			// |V|=89	
			break;
		}
		case 7:
		{
			Seed = 1227;
			init = 7;
			iter = 26;
			length = 7;
			// |V|=90	
			break;
		}
		case 8:
		{
			Seed = 1254;
			init = 7;
			iter = 22;
			length = 8;
			// |V|=91	
			break;
		}
		case 9:
		{
			Seed = 1267;
			init = 7;
			iter = 20;
			length = 9;
			// |V|=90
			break;
		}
		case 10:
		{
			Seed = 1267;
			init = 7;
			iter = 18;
			length = 10;
			// |V|=90	
			break;
		}
		case 11:
		{
			Seed = 1279;
			init = 7;
			iter = 16;
			length = 11;
			// |V|=91	
			break;
		}
		case 12:
		{
			Seed = 1301;
			init = 7;
			iter = 14;
			length = 12;
			// |V|=91	
			break;
		}
		case 13:
		{
			Seed = 1306;
			init = 7;
			iter = 13;
			length = 13;
			// |V|=91	
			break;
		}
		case 14:
		{
			Seed = 1317;
			init = 9;
			iter = 12;
			length = 14;
			// |V|=91	
			break;
		}
		case 15:
		{		
			Seed = 1323;
			init = 9;
			iter = 10;
			length = 15;
			// |V|=91	
			break;
		}
		case 16:
		{
			Seed = 1342;
			init = 11;
			iter = 9;
			length = 16;
			// |V|=89	
			break;
		}
		default:
		{
			break;
		}
		}

		char prefix[1024];
		char suffix[1024];
		char filename[1024];

		switch (method_type)
		{
		case BIBOX:
		{
			sprintf(prefix, "bibox_");
			break;
		}
		case BIBOX_MIT:
		{
			sprintf(prefix, "bibox_mit_");
			break;
		}
		case BIBOX_THETA_2:
		{
			sprintf(prefix, "bibox_theta_2_");
			break;
		}
		case BIBOX_THETA_3:
		{
			sprintf(prefix, "bibox_theta_3_");
			break;
		}
		case MIT:
		{
			sprintf(prefix, "mit_");
			break;
		}
		default:
		{
			break;
		}
		}

		switch (shortening_type)
		{
		case RAW:
		{
			sprintf(suffix, "biconnected_erasing_%d-%d-%d_raw.txt", init, iter, length);
			break;
		}
		case INVERSE:
		{
			sprintf(suffix, "biconnected_erasing_%d-%d-%d_inverse.txt", init, iter, length);
			break;
		}
		case REDUNDANT:
		{
			sprintf(suffix, "biconnected_erasing_%d-%d-%d_redundant.txt", init, iter, length);
			break;
		}
		case LONG:
		{
			sprintf(suffix, "biconnected_erasing_%d-%d-%d_long.txt", init, iter, length);
			break;
		}
		default:
		{
			break;
		}
		}

		FILE *fw;
		sprintf(filename, "%s%s", prefix, suffix);

		fw = fopen(filename, "w");

		if (fw == NULL)
		{
			printf("Cannot open %s. Terminating...\n", filename);
			exit(-1);
		}
	       
		for (free = 2; free <= 100; free += 1)
		{		
			fprintf(fw, "ID:%d (%d,%d,%d)\n", ID, iter, length, free);
			++ID;

			SOLUTION_DATABASE database;
			FINE_SOLUTION solution, filtered_solution;

			MOVE_COUNT = 0;
			STEP_COUNT = 0;
			GRAPH g, h, ch, hv;		       
					
			srandom(Seed);
			generate_biconnected(init, iter, length, free, &g);

			if (free < 2)
			{
				free = 2;
			}
			
			srandom(Seed);
			generate_biconnected(init, iter, length, free, &h);
		
			srandom(Seed);
			generate_biconnected(init, iter, length, free, &ch);
		
			srandom(Seed);
			generate_biconnected(init, iter, length, free, &hv);
		
			fprintf(fw, "Graph |V|:%d\n", g.m_vertices.size());
		
			if (h.m_vertices.size() > free)
			{
				begin_1 = clock();

				bool solved;
				
				switch (method_type)
				{
				case BIBOX:
				{
					solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
					break;
				}
				case BIBOX_MIT:
				{
					solved = solve_slider_bibox_theta(&h, database, solution, true, false, true);
					break;
				}
				case BIBOX_THETA_2:
				{
					solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
					break;
				}
				case BIBOX_THETA_3:
				{
					solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
					break;
				}
				case MIT:
				{
					solved = solve_slider_obox_theta(&h, database, solution, false, false);
					break;
				}
				default:
				{
					solved = false;
					break;
				}
				}
// Bibox_theta 2
//			bool solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
// Bibox_theta 3
//			bool solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
// Bibox
//			bool solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
			
				end_1 = clock();
				begin_2 = clock();
				
				fprintf(fw, "Solution size:%d\n", solution.size());

				filter_fine_solution(&ch, solution, filtered_solution);

				switch (shortening_type)
				{
				case RAW:
				{
					break;
				}
				case INVERSE:
				{
					erase_trivial_inverses(filtered_solution);
					break;
				}
				case REDUNDANT:
				{
					erase_independent_inverses(filtered_solution);
					break;
				}
				case LONG:
				{
					shorten_solution(filtered_solution, &hv);
					break;
				}
				default:
				{
					break;
				}
				}

//		erase_trivial_inverses(filtered_solution);
//		erase_complex_inverses(filtered_solution);
//		erase_long_inverses(filtered_solution);
//		erase_independent_inverses(filtered_solution);
//		shorten_solution(filtered_solution, &hv);
			
				parallelize_fine_solution_multirobotic(filtered_solution);

				end_2 = clock();
			
				fprintf(fw, "Solution size:%d\n", solution.size());
				fprintf(fw, "Filtered solution size:%d\n", filtered_solution.size());
				fprintf(fw, "Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
				fprintf(fw, "Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
				fprint_step_parallelism(fw, filtered_solution);
			//					print_fine_solution(filtered_solution);			

				fprintf(fw, "Solving time:%.3f\n", (end_1 - begin_1) / (double)CLOCKS_PER_SEC);
				fprintf(fw, "Filtering time:%.3f\n", (end_2 - begin_2) / (double)CLOCKS_PER_SEC);
			}
			fflush(fw);
		}
		fclose(fw);
	}
}


int main(int UNUSED(argc), char **UNUSED(argv))
{
	clock_t begin, end;

	printf("Mulirobot %s\n", VERSION);
	printf("(c) 2008-2010 %s\n\n", AUTHOR);

//	test_exchange_grids();

	test_grid_shortening(BIBOX_THETA_2, RAW);
	test_grid_shortening(BIBOX_THETA_2, INVERSE);
	test_grid_shortening(BIBOX_THETA_2, REDUNDANT);
	test_grid_shortening(BIBOX_THETA_2, LONG);

/*
	test_grid_shortening(MIT, RAW);
	test_grid_shortening(MIT, INVERSE);
	test_grid_shortening(MIT, REDUNDANT);
	test_grid_shortening(MIT, LONG);
*/
/*
	test_biconnected_shortening(BIBOX_THETA_2, RAW);
	test_biconnected_shortening(BIBOX_THETA_2, INVERSE);
	test_biconnected_shortening(BIBOX_THETA_2, REDUNDANT);
	test_biconnected_shortening(BIBOX_THETA_2, LONG);
*/
/*
	test_biconnected_shortening(MIT, RAW);
	test_biconnected_shortening(MIT, INVERSE);
	test_biconnected_shortening(MIT, REDUNDANT);
	test_biconnected_shortening(MIT, LONG);
*/
}
